from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestBallDeviceManualWithTarget(MpfTestCase):
    def __init__(self, test_map):
        super().__init__(test_map)
        self._captured = 0
        self._enter = 0
        self._missing = 0
        self._requesting = 0
        self._queue = False

    def get_config_file(self):
        return 'test_ball_device_manual_with_target.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def _missing_ball(self, **kwargs):
        del kwargs
        self._missing += 1

    def _requesting_ball(self, balls, **kwargs):
        del kwargs
        self._requesting += balls

    def _ball_enter(self, new_balls, unclaimed_balls, **kwargs):
        del kwargs
        del unclaimed_balls
        self._enter += new_balls

    def _captured_from_pf(self, balls, **kwargs):
        del kwargs
        self._captured += balls

    def test_manual_successful_eject_to_pf(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        self.assertEqual(1, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # request a ball
        playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)

        # launcher receives but waits for player to eject
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        # player shoots the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)

        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

    def test_manual_with_retry_to_pf(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)

        # trough should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        self.assertEqual(1, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # request a ball to the playfield
        playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(.1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)

        self.assertEqual(0, device1.balls)

        # launcher receives but waits for player to eject
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        # player shoots the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(0.1)

        # too soft and it comes back
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(10)
        self.assertEqual(1, device2.balls)
        assert not coil2.pulse.called

        # player drinks his coffee
        self.advance_time_and_run(300)

        # player shoots the ball again
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)
        assert not coil2.pulse.called

        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)

        self.assertEqual(1, playfield.balls)

        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

    def test_trough_retry(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        self.assertEqual(1, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # request an ball
        playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(0.1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)

        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()

        # ball falls back
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # it retries after a timeout
        self.advance_time_and_run(2)
        self.assertEqual(1, device1.balls)
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        # trough ejects again
        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)

        # launcher receives but waits for player to eject
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        # player shoots the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(0.1)

        # too soft and it comes back
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(11)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        # player drinks his coffee
        self.advance_time_and_run(300)

        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        # player shoots the ball again
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)

        self.assertEqual(1, playfield.balls)

        self.advance_time_and_run(100)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)
        self.assertEqual(1, self.machine.ball_controller.num_balls_known)

    def test_manual_fast_skipping_successful_eject_to_pf(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # request an ball
        playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()

        # playfield but count is not increased because eject_timeout of trough did not expire yet
        self.hit_and_release_switch("s_playfield")
        self.assertEqual(0, playfield.balls)

        # launcher does not see the ball. player ejects it right away. timeout expires
        self.advance_time_and_run(3)

        # ball hits the playfield (again)
        self.hit_and_release_switch("s_playfield")
        self.advance_time_and_run(1)

        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

        self.advance_time_and_run(100)
        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)
        self.assertEqual("idle", device1._state)
        self.assertEqual("idle", device2._state)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        self.assertEqual(1, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # will request a second ball. launcher has to use count eject confirmation
        playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)

        # launcher does not see the ball. player ejects it right away
        self.advance_time_and_run(1)

        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()

        self.assertEqual(1, playfield.balls)
        # since it will use count as eject confirm we have to wait for eject_timout of both devices
        self.advance_time_and_run(6 + 3)

        self.assertEqual(2, playfield.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

        self.advance_time_and_run(100)
        self.assertEqual(2, playfield.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)
        self.assertEqual("idle", device1._state)
        self.assertEqual("idle", device2._state)

    def test_capture_random_ball(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        device2 = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # launcher receives a random ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil2.pulse.called)

        # launcher should eject
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)

        self.assertEqual(1, playfield.balls)
        self.assertEqual(1, self._captured)
        self.assertEqual(0, self._missing)

    def test_manual_ball_missing(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        target = self.machine.ball_devices['test_target']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        self.assertEqual(1, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # request an ball
        playfield.add_ball(source_device=target, player_controlled=True)
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.assertEqual(playfield, target.outgoing_balls_handler._current_target)
        self.assertEqual(0, len(target.incoming_balls_handler._incoming_balls))

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)

        self.assertEqual(playfield, target.outgoing_balls_handler._current_target)
        self.advance_time_and_run(10)

        self.assertEqual(playfield, target.outgoing_balls_handler._current_target)

        # it does not hit any playfield switches and goes missing
        self.advance_time_and_run(100)

        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(1, self._missing)
        self.assertEqual("idle", device1._state)
        self.assertEqual("idle", device2._state)
        self.assertEqual("idle", target._state)

    def test_trough_eject_failed_with_manual(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add two initial balls to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # request an ball
        playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(0.1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        # and it comes back
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        # wait until timeout reached
        self.advance_time_and_run(2)
        self.assertEqual(2, device1.balls)

        # trough ejects again
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        # launcher receives but waits for player to eject
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)
        self.assertEqual("idle", device1._state)

        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # player shoots the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        self.advance_time_and_run(10)

        assert not coil1.pulse.called
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)

        self.assertEqual(1, playfield.balls)

        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

        self.advance_time_and_run(100)
        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

        assert not coil1.pulse.called
        assert not coil2.pulse.called

        self.assertEqual("idle", device1._state)
        self.assertEqual("idle", device2._state)

    def test_manual_successful_eject_with_no_pf_switch_hit(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add two initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # request an ball
        playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        # launcher receives but waits for player to eject
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        # player shoots the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        # it hits the pf but no pf switch (not confirmed yet)
        self.advance_time_and_run(1)
        #        self.assertEquals(0, playfield.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

        # it drains
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)

        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, self._captured)
        self.assertEqual(0, self._missing)

        self.advance_time_and_run(100)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, self._captured)
        self.assertEqual(0, self._missing)

    #    def _launcher_eject_attempt(self, balls, **kwargs):
    #        self._launcher_eject_attempt += balls

    def test_request_to_pf_and_launcher_and_unexpected_manual_eject(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add two initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # request an ball to pf
        playfield.add_ball(player_controlled=False)
        self.advance_time_and_run(1)

        # request an ball to launcher
        device2.request_ball()

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        # launcher receives
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()

        # launcher shoots the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)

        # launcher receives and keeps ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)
        # launcher should be idle
        self.assertEqual("idle", device2._state)

        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        self.advance_time_and_run(100)

        self.assertEqual(1, playfield.balls)
        self.assertEqual(1, playfield.available_balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

        # since we have a mechanical plunger the player decides to eject the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, self._captured)

        # both balls drain (before confirm of the second)
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, playfield.available_balls)
        self.assertEqual(2, self._captured)
        self.assertEqual(0, self._missing)

        self.advance_time_and_run(100)
        self.assertEqual(0, self._missing)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, playfield.available_balls)
        self.assertEqual(2, self._captured)
        # launcher should be idle
        self.assertEqual("idle", device2._state)

    def test_request_to_launcher_and_pf(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add two initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # request an ball to launcher
        device2.request_ball()

        # request an ball to pf
        playfield.add_ball(player_controlled=False)
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        # launcher receives
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()

        # launcher shoots the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)

        # thats it. no more ejects.
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        self.advance_time_and_run(100)
        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

    def test_request_launcher_with_manual_eject_and_skip(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add two initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # request an ball to launcher
        device2.request_ball()
        self.advance_time_and_run(0.1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        # it skips launcher and goes to pf
        self.advance_time_and_run(3)

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)

        self.assertEqual(1, playfield.balls)
        self.assertEqual(1, playfield.available_balls)

        # no launcher eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        self.advance_time_and_run(100)

        self.assertEqual(1, playfield.balls)
        self.assertEqual(1, playfield.available_balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)
        self.assertEqual("idle", device2._state)

    def test_request_manual_when_launcher_has_ball(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add two initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, playfield.available_balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # request an ball to launcher
        device2.request_ball()
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        # launcher receives
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        # ball stays in launcher
        self.assertTrue(coil1.pulse.called)
        coil1.pulse = MagicMock()
        assert not coil2.pulse.called

        # request an ball to pf
        playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        # still no eject. waiting for player
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # player shoots the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)

        # thats it. no more ejects.
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        self.advance_time_and_run(100)
        self.assertEqual(1, playfield.balls)
        self.assertEqual(1, playfield.available_balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

        self.assertEqual("idle", device1._state)
        self.assertEqual("idle", device2._state)

    def test_launcher_without_auto_fire_on_unexpected_ball(self):
        coil4 = self.machine.coils['eject_coil4']
        launcher_manual = self.machine.ball_devices['test_launcher_manual_on_unexpected']
        playfield = self.machine.ball_devices['playfield']

        # add ball to pf
        self.machine.ball_controller.num_balls_known = 1
        playfield.balls = 1
        playfield.available_balls = 1
        self.assertEqual(1, playfield.balls)
        coil4.pulse = MagicMock()
        self.assertEqual(0, launcher_manual.balls)
        assert not coil4.pulse.called

        # ball enters the launcher
        self.machine.switch_controller.process_switch("s_ball_switch_launcher2", 1)
        self.advance_time_and_run(1)

        # it should stay there and wait for manual eject
        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, launcher_manual.balls)
        self.assertEqual("ejecting", launcher_manual._state)
        assert not coil4.pulse.called

        # player has time
        self.advance_time_and_run(100)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, launcher_manual.balls)
        self.assertEqual("ejecting", launcher_manual._state)
        assert not coil4.pulse.called

        # but finally he ejects the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher2", 0)
        self.advance_time_and_run(1)
        self.assertEqual("ball_left", launcher_manual._state)
        self.advance_time_and_run(20)
        self.assertEqual("idle", launcher_manual._state)
        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, launcher_manual.balls)

    def test_vuk_with_auto_fire_on_unexpected_ball_false(self):
        coil5 = self.machine.coils['eject_coil5']
        vuk = self.machine.ball_devices['test_vuk']
        coil2 = self.machine.coils['eject_coil2']
        launcher = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        # add ball to pf
        self.machine.ball_controller.num_balls_known = 1
        playfield.balls = 1
        playfield.available_balls = 1
        coil5.pulse = MagicMock()
        coil2.pulse = MagicMock()
        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, launcher.balls)
        self.assertEqual(0, vuk.balls)
        assert not coil5.pulse.called
        assert not coil2.pulse.called

        # ball enters launcher unexpectedly
        self.machine.switch_controller.process_switch("s_ball_switch_launcher", 1)
        self.advance_time_and_run(1)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, launcher.balls)

        # launcher auto fires
        self.assertTrue(coil2.pulse.called)
        coil2.pulse = MagicMock()

        # ball leaves
        self.machine.switch_controller.process_switch("s_ball_switch_launcher", 0)

        # ball is back on pf
        self.advance_time_and_run(10)
        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, launcher.balls)

        # ball enters the vuk
        self.machine.switch_controller.process_switch("s_vuk", 1)
        self.advance_time_and_run(1)

        # vuk ejects to launcher
        self.assertTrue(coil5.pulse.called)
        coil5.pulse = MagicMock()

        # ball leaves
        self.machine.switch_controller.process_switch("s_vuk", 0)
        self.advance_time_and_run(1)

        # and enters launcher
        self.machine.switch_controller.process_switch("s_ball_switch_launcher", 1)
        self.advance_time_and_run(1)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, launcher.balls)
        self.assertEqual(0, vuk.balls)

        # it should stay there and wait for manual eject
        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, launcher.balls)
        self.assertEqual("ejecting", launcher._state)
        assert not coil2.pulse.called

        # player has time
        self.advance_time_and_run(100)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, launcher.balls)
        self.assertEqual("ejecting", launcher._state)
        assert not coil2.pulse.called

        # but finally he ejects the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher", 0)
        self.advance_time_and_run(1)
        self.assertEqual("ball_left", launcher._state)
        self.advance_time_and_run(20)
        self.assertEqual("idle", launcher._state)
        self.assertEqual("idle", vuk._state)
        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, launcher.balls)
        self.assertEqual(0, vuk.balls)



"""Command to format yaml files."""
import sys

import argparse
import re

from mpf.file_interfaces.yaml_roundtrip import YamlRoundtrip

from mpf.commands import MpfCommandLineParser
from mpf.tests.MpfDocTestCase import MpfDocTestCaseBase

SUBCOMMAND = True


class Command(MpfCommandLineParser):

    """Run a text unit test from cli."""

    def __init__(self, args, path):
        """Parse args."""
        super().__init__(args, path)
        test_file = self.argv.pop(1)

        parser = argparse.ArgumentParser(description='MPF Command')

        parser.add_argument("--yes", help="perform the action",
                            default=False, action="store_true", dest="do_it")

        args = parser.parse_args(self.argv[1:])

        with open(test_file) as f:
            test_string = f.read()

        if ".. code-block:: mpf-config" in test_string or ".. code-block:: mpf-mc-config" in test_string:
            print("Parsing documentation page {}.".format(test_file))
            new_config = re.sub(".. code-block:: (?P<type>mpf-config|mpf-mc-config)\n\n(?P<code>( {2,4}[^\n]+|\n)+)",
                                self._replace_config, test_string)
        elif test_file.endswith(".rst"):
            print("Rst file {} without test. Skipping.".format(test_file))
            sys.exit(3)
        else:
            print("Parsing single test {}.".format(test_file))
            new_config = self._reformat_test_case(test_string)

        if new_config != test_string:
            print("Config is not linted.")
            print(MpfDocTestCaseBase.unidiff_output(test_string, new_config))
            if args.do_it:
                print("Writing back changes")
                with open(test_file, "w") as f:
                    f.write(new_config)
                sys.exit(1)
            else:
                print("Not writing back changes. Use --yes to do this.")
                sys.exit(2)
        else:
            print("Config is ok.")
            sys.exit(0)

    def _replace_config(self, test_case):
        text = test_case.group("code")
        indent_len = None
        unindented_text = ''
        for line in text.split("\n"):
            if indent_len is None:
                indent = re.search("^( )+", line)
                if not indent:
                    raise AssertionError("Block is incorrectly indented: {}.".format(text))
                indent_len = len(indent.group(0))
            unindented_text += line[indent_len:] + "\n"

        undented_result = self._reformat_test_case(unindented_text)

        indented_result = ".. code-block:: " + test_case.group("type") + "\n\n"
        for line in undented_result.split("\n"):
            if line:
                indented_result += " " * indent_len + line + "\n"
            else:
                indented_result += "\n"
        return indented_result

    # pylint: disable-msg=too-many-locals
    def _reformat_test_case(self, test_case):
        test_case = test_case.replace("\t", "  ")
        test_case = re.sub(r" +\n", "\n", test_case)
        machine_config, extra_configs, mode_configs, show_configs, assets, code, tests = \
            MpfDocTestCaseBase.prepare_config(test_case, fixup_config=False)
        formatted_yaml = ""
        if machine_config:
            new_config = self._reformat(machine_config, show_file=False)
            if new_config == "null\n...\n":
                formatted_yaml += machine_config
            else:
                formatted_yaml += new_config
        for config_name, extra_config in extra_configs.items():
            formatted_yaml += "##! config: " + config_name + "\n"
            formatted_yaml += self._reformat(extra_config, show_file=False)
        for mode_name, mode_config in mode_configs.items():
            formatted_yaml += "##! mode: " + mode_name + "\n"
            formatted_yaml += self._reformat(mode_config, show_file=False)
        if show_configs:
            for show_name, show_config in show_configs.items():
                formatted_yaml += "##! show: " + show_name + "\n"
                formatted_yaml += self._reformat(show_config, show_file=True)

        if code:
            for file_path, file_content in code:
                formatted_yaml += "##! code: " + "/".join(file_path) + "\n"
                formatted_yaml += file_content

        if assets:
            for asset_path, asset_source in assets.items():
                formatted_yaml += "##! asset: {}={}\n".format(asset_path, asset_source)

        if tests:
            formatted_yaml += "##! test\n"
            for test in tests:
                formatted_yaml += test + "\n"

        return formatted_yaml.strip() + "\n"

    @staticmethod
    def _reformat(yaml_string, show_file):
        new_config = YamlRoundtrip.reformat_yaml(yaml_string, show_file)
        if new_config == "null\n...\n":
            return yaml_string

        return new_config



from unittest.mock import MagicMock

from mpf.tests.MpfTestCase import MpfTestCase, test_config


class TestCreditsMode(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/credits/'

    def start_game(self, should_work):
        # shots only work in games so we have to do this a lot
        self.machine.playfield.add_ball = MagicMock()
        self.machine.ball_controller.num_balls_known = 3
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        if should_work:
            self.assertIsNotNone(self.machine.game)
            self.machine.game.balls_in_play = 0
            self.advance_time_and_run()
        else:
            self.assertIsNone(self.machine.game)

    def start_two_player_game(self):
        # game start should work
        self.machine.playfield.add_ball = MagicMock()
        self.machine.ball_controller.num_balls_known = 3
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertIsNotNone(self.machine.game)
        self.assertEqual(1, self.machine.game.num_players)

        # add another player
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run(1)
        self.assertEqual(2, self.machine.game.num_players)

    def stop_game(self):
        # stop game
        self.assertIsNotNone(self.machine.game)
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertIsNone(self.machine.game)

    @test_config("config_freeplay.yaml")
    def test_free_play_at_start(self):
        self.assertEqual("FREE PLAY", self.machine.variables.get_machine_var('credits_string'))
        self.assertFalse(self.machine.variables.is_machine_var("price_per_game_raw_0"))
        self.assertFalse(self.machine.variables.is_machine_var("price_per_game_string_0"))

        self.start_two_player_game()

    def testToggleEvents(self):
        self.assertTrue(self.machine.mode_controller.is_active('credits'))
        self.assertEqual("CREDITS 0", self.machine.variables.get_machine_var('credits_string'))

        self.post_event("toggle_credit_play")
        self.assertEqual("FREE PLAY", self.machine.variables.get_machine_var('credits_string'))

        self.post_event("toggle_credit_play")
        self.assertEqual("CREDITS 0", self.machine.variables.get_machine_var('credits_string'))

        self.start_game(False)

        self.post_event("toggle_credit_play")
        self.assertEqual("FREE PLAY", self.machine.variables.get_machine_var('credits_string'))

        self.start_two_player_game()
        self.stop_game()

        self.post_event("enable_credit_play")
        self.assertEqual("CREDITS 0", self.machine.variables.get_machine_var('credits_string'))

        self.post_event("enable_credit_play")
        self.assertEqual("CREDITS 0", self.machine.variables.get_machine_var('credits_string'))

        self.post_event("enable_free_play")
        self.assertEqual("FREE PLAY", self.machine.variables.get_machine_var('credits_string'))

        self.post_event("enable_free_play")
        self.assertEqual("FREE PLAY", self.machine.variables.get_machine_var('credits_string'))

    def testCredits(self):
        self.assertTrue(self.machine.mode_controller.is_active('credits'))
        self.assertEqual("CREDITS 0", self.machine.variables.get_machine_var('credits_string'))

        # no credits no game
        self.start_game(False)

        self.hit_and_release_switch("s_left_coin")
        self.machine_run()

        self.assertEqual("CREDITS 1/2", self.machine.variables.get_machine_var('credits_string'))
        self.assertEqual(0.25, self.machine.modes["credits"].earnings["money"]["total_value"])
        self.assertMachineVarEqual(1.0, "credit_units")

        self.assertMachineVarEqual(0.5, "price_per_game_raw_0")
        self.assertMachineVarEqual("1 CREDITS $0.5", "price_per_game_string_0")
        self.assertMachineVarEqual(2, "price_per_game_raw_1")
        self.assertMachineVarEqual("5 CREDITS $2.0", "price_per_game_string_1")

        # not enough credits. no game
        self.start_game(False)

        self.hit_and_release_switch("s_left_coin")
        self.machine_run()

        self.assertEqual("CREDITS 1", self.machine.variables.get_machine_var('credits_string'))
        self.assertEqual(0.5, self.machine.modes["credits"].earnings["money"]["total_value"])

        # one is enough for a game
        self.start_game(True)
        self.stop_game()

        self.assertEqual("CREDITS 0", self.machine.variables.get_machine_var('credits_string'))

        # but only one
        self.start_game(False)

        self.hit_and_release_switch("s_right_coin")
        self.machine_run()
        self.assertEqual("CREDITS 2", self.machine.variables.get_machine_var('credits_string'))
        self.assertEqual(1.5, self.machine.modes["credits"].earnings["money"]["total_value"])

        # no more price tier after game
        self.hit_and_release_switch("s_left_coin")
        self.hit_and_release_switch("s_left_coin")
        self.machine_run()
        self.assertEqual("CREDITS 3", self.machine.variables.get_machine_var('credits_string'))
        self.assertEqual(2.0, self.machine.modes["credits"].earnings["money"]["total_value"])

        self.post_event("earnings_reset")
        self.advance_time_and_run(.1)

        self.assertEqual({}, self.machine.modes["credits"].earnings)
        self.assertEqual({}, self.machine.modes["credits"].data_manager.written_data)

        self.assertMachineVarEqual(6.0, "credit_units")
        self.post_event("credits_reset")
        self.advance_time_and_run(.1)
        self.assertEqual("CREDITS 0", self.machine.variables.get_machine_var('credits_string'))
        self.assertMachineVarEqual(0.0, "credit_units")

        # should not be able to start a game without credits
        self.start_game(False)

    def testReplay(self):
        # add coins
        self.hit_and_release_switch("s_left_coin")
        self.hit_and_release_switch("s_left_coin")
        self.advance_time_and_run()
        self.assertEqual("CREDITS 1", self.machine.variables.get_machine_var('credits_string'))
        # start game
        self.start_game(True)

        self.assertEqual("CREDITS 0", self.machine.variables.get_machine_var('credits_string'))
        # no replay
        self.stop_game()

        # try again
        self.hit_and_release_switch("s_left_coin")
        self.hit_and_release_switch("s_left_coin")
        self.advance_time_and_run()
        self.assertEqual("CREDITS 1", self.machine.variables.get_machine_var('credits_string'))
        self.start_game(True)

        # score 600k
        self.machine.game.player.score = 600000

        # replay credit on game end
        self.stop_game()
        self.assertEqual("CREDITS 1", self.machine.variables.get_machine_var('credits_string'))

    def testMorePlayers(self):
        self.assertTrue(self.machine.mode_controller.is_active('credits'))
        self.assertEqual("CREDITS 0", self.machine.variables.get_machine_var('credits_string'))

        self.hit_and_release_switch("s_left_coin")
        self.hit_and_release_switch("s_left_coin")
        self.machine_run()

        self.assertEqual("CREDITS 1", self.machine.variables.get_machine_var('credits_string'))

        # one is enough for a game
        self.machine.playfield.add_ball = MagicMock()
        self.machine.ball_controller.num_balls_known = 3
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertIsNotNone(self.machine.game)

        # no more credits
        self.assertEqual("CREDITS 0", self.machine.variables.get_machine_var('credits_string'))
        self.assertEqual(1, self.machine.game.num_players)
        # try to add another player
        self.hit_and_release_switch("s_start")
        # fails
        self.assertEqual(1, self.machine.game.num_players)

        # add credits
        self.hit_and_release_switch("s_right_coin")
        self.machine_run()
        self.assertEqual("CREDITS 2", self.machine.variables.get_machine_var('credits_string'))

        # try to add another player
        self.hit_and_release_switch("s_start")
        # wrorks
        self.assertEqual(2, self.machine.game.num_players)
        self.machine_run()
        self.assertEqual("CREDITS 1", self.machine.variables.get_machine_var('credits_string'))

    def testMaxCredits(self):
        self.assertTrue(self.machine.mode_controller.is_active('credits'))
        self.assertEqual("CREDITS 0", self.machine.variables.get_machine_var('credits_string'))

        self.hit_and_release_switch("s_right_coin")
        self.hit_and_release_switch("s_right_coin")
        self.hit_and_release_switch("s_right_coin")
        self.hit_and_release_switch("s_right_coin")
        self.machine_run()

        self.assertEqual("CREDITS 10", self.machine.variables.get_machine_var('credits_string'))

        self.hit_and_release_switch("s_right_coin")
        self.machine_run()
        self.assertEqual("CREDITS 12", self.machine.variables.get_machine_var('credits_string'))

        self.hit_and_release_switch("s_right_coin")
        self.machine_run()
        self.assertEqual("CREDITS 12", self.machine.variables.get_machine_var('credits_string'))

    def testPricingTiers(self):
        self.hit_and_release_switch("s_right_coin")
        self.machine_run()
        self.assertEqual("CREDITS 2", self.machine.variables.get_machine_var('credits_string'))

        self.hit_and_release_switch("s_right_coin")
        self.machine_run()
        self.assertEqual("CREDITS 5", self.machine.variables.get_machine_var('credits_string'))

    def testFractionalTimeout(self):
        self.hit_and_release_switch("s_right_coin")
        self.hit_and_release_switch("s_left_coin")
        self.machine_run()
        self.assertEqual("CREDITS 2 1/2", self.machine.variables.get_machine_var('credits_string'))

        self.advance_time_and_run(60 * 15)

        self.assertEqual("CREDITS 2", self.machine.variables.get_machine_var('credits_string'))

        # but not during game
        self.hit_and_release_switch("s_left_coin")
        self.machine_run()
        self.assertEqual("CREDITS 2 1/2", self.machine.variables.get_machine_var('credits_string'))

        self.start_game(True)
        self.advance_time_and_run(60 * 15)
        self.stop_game()

        self.machine_run()
        self.assertEqual("CREDITS 1 1/2", self.machine.variables.get_machine_var('credits_string'))

        # but timeout restarts
        self.advance_time_and_run(60 * 15)

        self.assertEqual("CREDITS 1", self.machine.variables.get_machine_var('credits_string'))

    def testCreditTimeout(self):
        self.hit_and_release_switch("s_right_coin")
        self.hit_and_release_switch("s_left_coin")
        self.machine_run()
        self.assertEqual("CREDITS 2 1/2", self.machine.variables.get_machine_var('credits_string'))

        self.advance_time_and_run(3600 * 2)

        self.assertEqual("CREDITS 0", self.machine.variables.get_machine_var('credits_string'))

        # but not during game
        self.hit_and_release_switch("s_right_coin")
        self.hit_and_release_switch("s_left_coin")
        self.machine_run()
        self.assertEqual("CREDITS 2 1/2", self.machine.variables.get_machine_var('credits_string'))

        self.start_game(True)
        self.advance_time_and_run(3600 * 2)
        self.stop_game()

        self.machine_run()
        self.assertEqual("CREDITS 1 1/2", self.machine.variables.get_machine_var('credits_string'))

        # but timeout restarts
        self.advance_time_and_run(3600 * 2)

        self.assertEqual("CREDITS 0", self.machine.variables.get_machine_var('credits_string'))

    def testServiceCredits(self):
        self.hit_and_release_switch("s_esc")
        self.machine_run()
        self.assertEqual("CREDITS 1", self.machine.variables.get_machine_var('credits_string'))



from unittest.mock import MagicMock

from mpf.core.settings_controller import SettingEntry
from mpf.plugins.auditor import Auditor
from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase


class TestServiceMode(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/service_mode/'

    def setUp(self):
        self.machine_config_patches['mpf']['plugins'] = ['mpf.plugins.auditor.Auditor']
        super().setUp()

    def test_start_stop_service_in_attract(self):
        self.mock_event("service_door_opened")
        self.mock_event("service_door_closed")
        self.mock_event("service_mode_entered")
        self.mock_event("service_mode_exited")

        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run()
        self.assertMachineVarEqual(1, "credit_units")
        self.assertModeRunning("attract")

        # open door
        self.hit_switch_and_run("s_door_open", 1)
        self.assertEventCalled('service_door_opened', 1)
        self.assertEventNotCalled('service_door_closed')
        self.assertModeRunning("attract")

        # enter
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()
        self.assertEventCalled('service_mode_entered', 1)
        self.assertEventNotCalled('service_mode_exited')
        self.assertModeNotRunning("attract")

        # exit
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run()
        self.assertEventCalled('service_mode_entered', 1)
        self.assertEventCalled('service_mode_exited', 1)
        self.assertModeRunning("attract")
        self.assertModeRunning("service")

        # enter again
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()
        self.assertEventCalled('service_mode_entered', 2)
        self.assertEventCalled('service_mode_exited', 1)
        self.assertModeNotRunning("attract")

        # close door
        self.release_switch_and_run("s_door_open", 1)
        self.assertEventCalled('service_mode_exited', 1)
        self.assertEventCalled('service_door_closed', 1)
        self.assertModeNotRunning("attract")
        self.assertModeRunning("service")

        # exit service
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run()
        self.assertEventCalled('service_mode_exited', 2)
        self.assertEventCalled('service_door_closed', 1)
        self.assertModeRunning("attract")
        self.assertModeRunning("service")
        self.assertModeRunning("credits")

        self.assertMachineVarEqual(1, "credit_units")

        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run()
        self.assertMachineVarEqual(2, "credit_units")

    def test_start_stop_service_in_game(self):
        self.assertModeRunning("service")
        self.mock_event("service_door_opened")
        self.mock_event("service_door_closed")
        self.mock_event("service_mode_entered")
        self.mock_event("service_mode_exited")

        # add credit
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run()

        self.start_game()
        self.assertModeRunning("game")
        self.assertModeRunning("service")
        self.assertSwitchState("s_door_open", 0)

        # open door. game still running
        self.hit_switch_and_run("s_door_open", .1)
        self.assertEventCalled('service_door_opened')
        self.assertEventNotCalled('service_door_closed')
        self.assertModeRunning("game")
        self.assertModeRunning("service")

        # enter service. end game
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()
        self.assertEventCalled('service_mode_entered')
        self.assertEventNotCalled('service_mode_exited')
        self.assertModeNotRunning("game")

        # close door. service mode still active
        self.release_switch_and_run("s_door_open", 1)
        self.assertEventCalled('service_door_closed')
        self.assertModeNotRunning("attract")

        # exit service mode
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run()
        self.assertModeRunning("attract")
        self.assertEventCalled('service_mode_exited')

    def test_start_menu(self):
        self.mock_event("service_menu_selected")
        # enter menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Diagnostics Menu')

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_menu_selected", label='Audits Menu')

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_menu_selected", label='Adjustments Menu')

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_menu_selected", label='Utilities Menu')

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_menu_selected", label='Diagnostics Menu')

        self.hit_and_release_switch("s_service_down")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_menu_selected", label='Utilities Menu')

    def test_utilities_reset(self):
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run(.1)
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run(.1)
        self.assertMachineVarEqual(2, "credit_units")
        self.assertEqual({'service_credit': {'total_value': 2, 'count': 2}}, self.machine.modes["credits"].earnings)

        self.start_game()
        self.hit_and_release_switch("s_service_up")
        self.machine.game.player.score = 1000
        self.drain_all_balls()
        self.assertGameIsNotRunning()

        auditor = self.machine.plugins[0]
        self.assertIsInstance(auditor, Auditor)
        self.assertEqual(1, auditor.current_audits['switches']['s_service_up'])
        self.assertTrue(1000, auditor.current_audits['player']['score']['average'])
        self.assertTrue(1, auditor.current_audits['events']['game_ended'])

        self.assertMachineVarEqual(1, "credit_units")

        self.mock_event("service_menu_selected")
        self.mock_event("service_options_slide_start")
        # enter menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Diagnostics Menu')

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_menu_selected", label='Audits Menu')

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_menu_selected", label='Adjustments Menu')

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_menu_selected", label='Utilities Menu')

        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_menu_selected", label='Reset Menu')

        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_menu_selected", label='Reset Coin Audits')
        self.assertEventNotCalled("service_options_slide_start")
        self.mock_event("service_menu_selected")
        self.mock_event("service_options_slide_start")

        # RESET earning audits
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_options_slide_start", title="Reset earning audits",
                                   question="Perform coin reset?", option="no", warning="THIS CANNOT BE UNDONE")

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_options_slide_start", title="Reset earning audits",
                                   question="Perform coin reset?", option="yes", warning="THIS CANNOT BE UNDONE")
        self.assertEventNotCalled("service_menu_selected")
        self.mock_event("service_options_slide_start")

        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_menu_selected", label='Reset Coin Audits')
        self.assertEventNotCalled("service_options_slide_start")

        # coins are still there
        self.assertMachineVarEqual(1, "credit_units")
        # but earning audits are reset
        self.assertEqual({}, self.machine.modes["credits"].earnings)

        # RESET game audits
        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_menu_selected", label='Reset Game Audits')
        self.mock_event("service_menu_selected")
        self.mock_event("service_options_slide_start")

        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_options_slide_start", title="Auditor Reset",
                                   question="Reset Game Audits?", option="no", warning="THIS CANNOT BE UNDONE")

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_options_slide_start", title="Auditor Reset",
                                   question="Reset Game Audits?", option="yes", warning="THIS CANNOT BE UNDONE")
        self.assertEventNotCalled("service_menu_selected")
        self.mock_event("service_options_slide_start")

        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_menu_selected", label='Reset Game Audits')
        self.assertEventNotCalled("service_options_slide_start")

        self.assertEqual(0, auditor.current_audits['switches']['s_service_up'])
        self.assertEqual(0, auditor.current_audits['player']['score']['average'])
        self.assertEqual(0, auditor.current_audits['events']['game_ended'])

        # RESET high scores (TEST NOT IMPLEMENTED YET - FEEL FREE)
        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_menu_selected", label='Reset High Scores')
        self.mock_event("service_menu_selected")
        self.mock_event("service_options_slide_start")

        # RESET credits
        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_menu_selected", label='Reset Credits')
        self.mock_event("service_menu_selected")
        self.mock_event("service_options_slide_start")

        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_options_slide_start", title="Reset credits",
                                   question="Remove all credits?", option="no", warning="THIS CANNOT BE UNDONE")

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_options_slide_start", title="Reset credits",
                                   question="Remove all credits?", option="yes", warning="THIS CANNOT BE UNDONE")
        self.assertEventNotCalled("service_menu_selected")
        self.mock_event("service_options_slide_start")

        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_menu_selected", label='Reset Credits')
        self.assertEventNotCalled("service_options_slide_start")

        # credits should be gone
        self.assertMachineVarEqual(0, "credit_units")

    def test_switch_test(self):
        self.mock_event("service_menu_selected")
        # enter menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Diagnostics Menu')

        # enter diagnostics menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Switch Menu')

        # enter switch menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Switch Edge Test')

        # start edge test
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        self.mock_event("service_switch_test_start")
        self.mock_event("service_switch_test_stop")
        self.hit_switch_and_run("s_start", 1)

        self.assertEventCalledWith("service_switch_test_start", switch_label='%', switch_name='s_start',
                                   switch_num='', switch_state='active')

        # leave switch test
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run()
        self.assertEventCalled("service_switch_test_stop")

    def test_light_test(self):
        self.mock_event("service_menu_selected")
        # enter menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Diagnostics Menu')

        # enter diagnostics menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Switch Menu')

        # select coil menu
        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Coil Menu')

        # select light menu
        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Light Menu')

        # enter light menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Single Light Test')

        self.mock_event("service_light_test_start")
        self.mock_event("service_light_test_stop")

        # enter single light test
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        for color in ["white", "red", "green", "blue", "yellow", "white"]:
            self.assertEventCalledWith("service_light_test_start",
                                       board_name='Virtual',
                                       light_label='%',
                                       light_name='l_light1',
                                       light_num='1',
                                       test_color=color)

            self.assertLightColor("l_light1", color)

            self.hit_and_release_switch("s_service_enter")
            self.advance_time_and_run()

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_light_test_start",
                                   board_name='Virtual',
                                   light_label='%',
                                   light_name='l_light5',
                                   light_num='5',
                                   test_color="red")

        self.assertLightColor("l_light1", "black")
        self.assertLightColor("l_light5", "red")

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_light_test_start",
                                   board_name='Virtual',
                                   light_label='%',
                                   light_name='l_light1',
                                   light_num='1',
                                   test_color="red")

        self.assertLightColor("l_light1", "red")
        self.assertLightColor("l_light5", "black")

        # leave switch test
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run()
        self.assertEventCalled("service_light_test_stop")

    def test_coil_test(self):
        self.mock_event("service_menu_selected")
        self.mock_event("service_coil_test_start")
        self.mock_event("service_coil_test_stop")
        # enter menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Diagnostics Menu')

        # enter diagnostics menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Switch Menu')

        # select coil menu
        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Coil Menu')

        # enter coil menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Single Coil Test')

        # start edge test
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        # selects the first coil
        self.assertEventCalledWith("service_coil_test_start", board_name='Virtual', coil_label='First coil',
                                   coil_name='c_test', coil_num='1')

        # select test2
        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_coil_test_start", board_name='Virtual', coil_label='Second coil',
                                   coil_name='c_test2', coil_num='2')

        self.machine.coils["c_test2"].pulse = MagicMock()
        # pulse it
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()
        self.machine.coils["c_test2"].pulse.assert_called_with()

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_coil_test_start", board_name='Virtual', coil_label='Third coil',
                                   coil_name='c_test5', coil_num='3')

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_coil_test_start", board_name='Virtual', coil_label='Fourth coil',
                                   coil_name='c_test6', coil_num='10')

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_coil_test_start", board_name='Virtual', coil_label='Fifth coil',
                                   coil_name='c_test4', coil_num='100')

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_coil_test_start", board_name='Virtual', coil_label='Sixth coil',
                                   coil_name='c_test3', coil_num='1000')



        # wrap to first
        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()

        # selects the first coil
        self.assertEventCalledWith("service_coil_test_start", board_name='Virtual', coil_label='First coil',
                                   coil_name='c_test', coil_num='1')

        # wrap back to last
        self.hit_and_release_switch("s_service_down")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_coil_test_start", board_name='Virtual', coil_label='Sixth coil',
                                   coil_name='c_test3', coil_num='1000')

        # leave coil test
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run()
        self.assertEventCalled("service_coil_test_stop")

    def test_volume(self):
        self.mock_event("master_volume_increase")
        self.mock_event("master_volume_decrease")

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalled("master_volume_increase", 1)
        self.assertEventCalled("master_volume_decrease", 0)

        self.hit_and_release_switch("s_service_down")
        self.advance_time_and_run()
        self.assertEventCalled("master_volume_increase", 1)
        self.assertEventCalled("master_volume_decrease", 1)

    def test_settings(self):
        self.machine.settings._settings = {}
        self.machine.settings.add_setting(SettingEntry("test1", "Test1", 1, "test1", "b",
                                                       {"a": "A", "b": "B (default)", "c": "C"}))
        self.machine.settings.add_setting(SettingEntry("test2", "Test2", 2, "test2", False,
                                                       {True: "Yes", False: "No (default)"}))
        self.mock_event("service_settings_start")
        self.mock_event("service_settings_edit")
        self.mock_event("service_settings_stop")

        self.mock_event("service_menu_selected")
        # enter menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Diagnostics Menu')

        # select audits menu
        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Audits Menu')

        # select adjustments menu
        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Adjustments Menu')

        # enter adjustments menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_menu_selected", label='Standard Adjustments')

        # enter standard adjustments menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()

        self.assertEventCalledWith("service_settings_start", settings_label='Test1', value_label="B (default)")

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_settings_start", settings_label='Test2', value_label="No (default)")

        # change setting
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_settings_edit", settings_label='Test2', value_label="No (default)")

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_settings_edit", settings_label='Test2', value_label="Yes")

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_settings_edit", settings_label='Test2', value_label="No (default)")

        self.hit_and_release_switch("s_service_down")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_settings_edit", settings_label='Test2', value_label="Yes")

        # exit setting change
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run()

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_settings_start", settings_label='Test1', value_label="B (default)")

        self.hit_and_release_switch("s_service_down")
        self.advance_time_and_run()
        self.assertEventCalledWith("service_settings_start", settings_label='Test2', value_label="Yes")

        self.assertEventNotCalled("service_settings_stop")

        # exit settings change
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run()

        self.assertEventCalled("service_settings_stop")



"""Base class for serial communicator."""
from typing import Optional

import asyncio
from serial import SerialException

from mpf.core.utility_functions import Util


MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


class BaseSerialCommunicator:

    """Basic Serial Communcator for platforms."""

    __slots__ = ["machine", "platform", "log", "debug", "port", "baud", "xonxoff", "reader", "writer", "read_task"]

    # pylint: disable=too-many-arguments
    def __init__(self, platform, port: str, baud: int, xonxoff=False) -> None:
        """Initialise Serial Connection Hardware.

        Args:
        ----
            platform(mpf.core.platform.BasePlatform): the platform
            port: Port to open.
            baud: Baudrate to use
            xonxoff (bool): Use xonxoff as flow control
        """
        self.machine = platform.machine     # type: MachineController
        self.platform = platform
        self.log = self.platform.log
        self.debug = self.platform.config['debug']
        self.port = port
        self.baud = baud
        self.xonxoff = xonxoff
        self.reader = None      # type: Optional[asyncio.StreamReader]
        self.writer = None      # type: Optional[asyncio.StreamWriter]
        self.read_task = None

    async def connect(self):
        """Connect to the hardware."""
        await self._connect_to_hardware(self.port, self.baud, self.xonxoff)

    async def _connect_to_hardware(self, port, baud, xonxoff=False):
        self.log.info("Connecting to %s at %sbps", port, baud)
        while True:
            try:
                connector = self.machine.clock.open_serial_connection(
                    url=port, baudrate=baud, limit=0, xonxoff=xonxoff)
                self.reader, self.writer = await connector
            except SerialException:
                if not self.machine.options["production"]:
                    raise

                # if we are in production mode retry
                await asyncio.sleep(.1)
                self.log.debug("Connection to %s failed. Will retry.", port)
            else:
                # we got a connection
                break

        serial = self.writer.transport.serial
        if hasattr(serial, "set_low_latency_mode"):
            serial.set_low_latency_mode(True)

        # defaults are slightly high for our usecase
        self.writer.transport.set_write_buffer_limits(2048, 1024)

        # read everything which is sitting in the serial
        self.writer.transport.serial.reset_input_buffer()
        # clear buffer
        # pylint: disable-msg=protected-access
        self.reader._buffer = bytearray()

        await self._identify_connection()

    async def start_read_loop(self):
        """Start the read loop."""
        self.read_task = self.machine.clock.loop.create_task(self._socket_reader())
        self.read_task.add_done_callback(Util.raise_exceptions)

    # pylint: disable-msg=inconsistent-return-statements
    async def readuntil(self, separator, min_chars: int = 0):
        """Read until separator.

        Args:
        ----
            separator: Read until this separator byte.
            min_chars: Minimum message length before separator
        """
        assert self.reader is not None
        # asyncio StreamReader only supports this from python 3.5.2 on
        buffer = b''
        while True:
            char = await self.reader.readexactly(1)
            buffer += char
            if char == separator and len(buffer) > min_chars:
                if self.debug:
                    self.log.debug("%s received: %s (%s)", self, buffer, "".join(" 0x%02x" % b for b in buffer))
                return buffer

    async def read(self, n=-1):
        """Read up to `n` bytes from the stream and log the result if debug is true.

        See :func:`StreamReader.read` for details about read and the `n` parameter.
        """
        try:
            resp = await self.reader.read(n)
        except asyncio.CancelledError:  # pylint: disable-msg=try-except-raise
            raise
        except Exception as e:  # pylint: disable-msg=broad-except
            self.log.warning("Serial error: {}".format(e))
            return None

        # we either got empty response (-> socket closed) or and error
        if not resp:
            self.log.warning("Serial closed.")
            self.machine.stop("Serial {} closed.".format(self.port))
            return None

        if self.debug:
            self.log.debug("%s received: %s (%s)", self, resp, "".join(" 0x%02x" % b for b in resp))
        return resp

    async def _identify_connection(self):
        """Initialise and identify connection."""
        raise NotImplementedError("Implement!")

    def stop(self):
        """Stop and shut down this serial connection."""
        self.log.error("Stop called on serial connection %s", self.port)
        if self.read_task:
            self.read_task.cancel()
            self.read_task = None
        if self.writer:
            self.writer.close()
            self.writer = None

    def send(self, msg):
        """Send a message to the remote processor over the serial connection.

        Args:
        ----
            msg: Byes of the message you want to send.
        """
        if self.debug:
            self.log.debug("%s sending: %s (%s)", self, msg, "".join(" 0x%02x" % b for b in msg))
        self.writer.write(msg)

    def _parse_msg(self, msg):
        """Parse a message.

        Msg may be partial.

        Args:
        ----
            msg: Bytes of the message (part) received.
        """
        raise NotImplementedError("Implement!")

    def __repr__(self):
        """Return str representation."""
        return self.port

    async def _socket_reader(self):
        while True:
            resp = await self.read(128)
            if resp is None:
                return
            self._parse_msg(resp)



"""Test the BallController."""
from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestBallControllerRegression(MpfTestCase):

    def setUp(self):
        super().setUp()
        #self.machine.ball_controller.num_balls_known = 0

    def get_config_file(self):
        return 'regression.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_controller/'

    def get_platform(self):
        return 'virtual'

    def test_regression(self):
        self.release_switch_and_run("s_shooter_lane", 4)
        self.hit_switch_and_run("s_trough_5", .1)
        self.hit_switch_and_run("s_underRightRampEject", .1)
        self.release_switch_and_run("s_underRightRampEject", .2)
        self.hit_switch_and_run("s_trough_6", .2)
        self.hit_switch_and_run("s_underRightRampEject", .1)
        self.release_switch_and_run("s_underRightRampEject", .2)
        self.advance_time_and_run(10)
        self.assertEqual(6, self.machine.ball_controller.num_balls_known)
        self.assertBallsOnPlayfield(0)
        self.assertAvailableBallsOnPlayfield(0)

class TestBallController(MpfTestCase):
    def setUp(self):
        super().setUp()
        self.machine.ball_controller.num_balls_known = 0

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_controller/'

    def get_platform(self):
        return 'smart_virtual'

    def test_game_start(self):
        # min balls is set to 3
        self.assertEqual(None, self.machine.game)
        self.advance_time_and_run(1)
        self.assertEqual(None, self.machine.game)
        self.assertEqual(0, self.machine.ball_controller.num_balls_known)

        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        # not enough balls
        self.assertEqual(None, self.machine.game)
        self.machine.switch_controller.process_switch("s_start", 1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)
        self.assertEqual(None, self.machine.game)

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.machine.switch_controller.process_switch("s_ball_switch3", 1)
        self.machine.switch_controller.process_switch("s_ball_switch4", 1)
        self.advance_time_and_run(1)
        self.assertEqual(3, self.machine.ball_controller.num_balls_known)

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)

        # not all balls are home
        self.assertEqual(None, self.machine.game)
        self.machine.switch_controller.process_switch("s_start", 1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)
        self.assertEqual(None, self.machine.game)

        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(.01)
        # ball did not settle yet
        self.machine.switch_controller.process_switch("s_start", 1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)
        self.assertEqual(None, self.machine.game)

        self.assertEqual(4, self.machine.ball_controller.num_balls_known)

        self.assertEqual(None, self.machine.game)
        self.machine.switch_controller.process_switch("s_start", 1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)
        self.assertNotEqual(None, self.machine.game)

    def test_ball_collect(self):
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch3", 1)
        self.machine.switch_controller.process_switch("s_ball_switch4", 1)
        self.advance_time_and_run(1)
        self.assertEqual(3, self.machine.ball_controller.num_balls_known)

        # prevent smart_virtual from "ejecting" the ball
        self.machine.coils["eject_coil2"].pulse = MagicMock()

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)

        self.assertEqual("idle", self.machine.ball_devices["test_trough"]._state)
        self.assertEqual("ejecting", self.machine.ball_devices["test_launcher"]._state)

        # not all balls are home. it should trigger a collect
        self.machine.switch_controller.process_switch("s_start", 1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(.1)
        self.assertEqual(None, self.machine.game)

        # ball is still in launcher and the collect ball should not try to eject it again (which would request a ball
        # at the trough. so check that trough is idle
        self.assertEqual("idle", self.machine.ball_devices["test_trough"]._state)
        self.assertEqual("ejecting", self.machine.ball_devices["test_launcher"]._state)

    def test_ball_collect_after_game(self):
        self.mock_event("collecting_balls_complete")
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.machine.switch_controller.process_switch("s_ball_switch3", 1)
        self.machine.switch_controller.process_switch("s_ball_switch4", 1)
        self.advance_time_and_run(1)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)

        # eject one ball to launcher
        self.machine.ball_devices["test_trough"].eject(target=self.machine.ball_devices["test_launcher"])
        self.advance_time_and_run(1)
        self.assertEqual(3, self.machine.ball_devices["test_trough"].available_balls)
        self.assertEqual(1, self.machine.ball_devices["test_launcher"].available_balls)

        # assume the game ended and we want to collect all balls
        self.machine.ball_controller.collect_balls()
        self.advance_time_and_run(1)

        self.assertEqual("idle", self.machine.ball_devices["test_trough"]._state)
        self.assertEqual("ball_left", self.machine.ball_devices["test_launcher"]._state)
        self.assertEqual(3, self.machine.ball_devices["test_trough"].available_balls)
        self.assertEqual(0, self.machine.ball_devices["test_launcher"].available_balls)

        self.advance_time_and_run(11)

        # both devices should be idle
        self.assertEqual("idle", self.machine.ball_devices["test_trough"]._state)
        self.assertEqual("idle", self.machine.ball_devices["test_launcher"]._state)
        self.assertEqual(3, self.machine.ball_devices["test_trough"].available_balls)
        self.assertEqual(0, self.machine.ball_devices["test_launcher"].available_balls)
        self.assertEqual(3, self.machine.ball_devices["test_trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["test_launcher"].balls)
        self.assertEqual(0, self._events['collecting_balls_complete'])

        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["test_trough"])
        self.advance_time_and_run(1)

        self.assertEqual(1, self._events['collecting_balls_complete'])

    def test_ball_collect_without_balls_on_pf(self):
        self.mock_event("collecting_balls_complete")
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.machine.switch_controller.process_switch("s_ball_switch3", 1)
        self.machine.switch_controller.process_switch("s_ball_switch4", 1)
        self.advance_time_and_run(1)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)

        self.machine.ball_controller.collect_balls()
        self.advance_time_and_run(1)

        self.assertEqual(1, self._events['collecting_balls_complete'])

    def test_unknown_ball_on_playfield(self):
        self.mock_event("collecting_balls_complete")
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.machine.switch_controller.process_switch("s_ball_switch3", 1)
        self.machine.switch_controller.process_switch("s_ball_switch4", 0)
        self.advance_time_and_run(1)
        self.assertEqual(3, self.machine.ball_controller.num_balls_known)

        self.hit_and_release_switch("s_playfield")
        self.advance_time_and_run(1)
        self.assertEqual(3, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, self.machine.playfield.available_balls)
        self.assertEqual(0, self.machine.playfield.balls)

        self.machine.switch_controller.process_switch("s_ball_switch4", 1)
        self.advance_time_and_run(1)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)

    def test_loose_balls(self):
        self.mock_event("collecting_balls_complete")
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.machine.switch_controller.process_switch("s_ball_switch3", 1)
        self.machine.switch_controller.process_switch("s_ball_switch4", 1)
        self.advance_time_and_run(1)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)

        self.machine.switch_controller.process_switch("s_ball_switch3", 0)
        self.machine.switch_controller.process_switch("s_ball_switch4", 0)
        self.advance_time_and_run(1)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)

        self.machine.switch_controller.process_switch("s_ball_switch4", 1)
        self.advance_time_and_run(1)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)



from mpf.tests.MpfGameTestCase import MpfGameTestCase
from mpf.tests.MpfTestCase import test_config, test_config_directory


class TestTilt(MpfGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/tilt/'

    def get_platform(self):
        return 'smart_virtual'

    def _tilted(self, **kwargs):
        del kwargs
        self._is_tilted = True

    def _prepare_trough(self):
        self.machine.ball_controller.num_balls_known = 0
        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.machine.switch_controller.process_switch('s_ball_switch2', 1)
        self.advance_time_and_run(2)
        self.assertEqual(2, self.machine.ball_devices["bd_trough"].balls)

    def _add_tilt_handler(self):
        self._is_tilted = False
        self.machine.events.add_handler("tilt", self._tilted)

    @test_config("config_system_11_trough.yaml")
    def test_tilt_in_outhole(self):
        """Test that the ball does not stay in the outhole."""
        self._add_tilt_handler()

        self._prepare_trough()
        self.start_game(2)
        self.advance_time_and_run(10)

        self.assertBallsOnPlayfield(0)
        self.assertAvailableBallsOnPlayfield(1)
        self.assertEqual(1, self.machine.ball_devices["bd_plunger"].balls)

        # ball ejects and ends up on playfield
        self.release_switch_and_run("s_plunger", 10)
        self.assertBallsOnPlayfield(1)
        self.assertAvailableBallsOnPlayfield(1)

        # machine tilts
        self.assertFalse(self._is_tilted)
        self.hit_and_release_switch("s_tilt")
        self.advance_time_and_run(10)
        self.assertTrue(self._is_tilted)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(True, self.machine.game.tilted)
        self.assertEqual(0, self.machine.ball_devices["bd_outhole"].balls)
        self.assertEqual(0, self.machine.ball_devices["bd_plunger"].balls)
        self.assertBallsOnPlayfield(1)
        self.assertAvailableBallsOnPlayfield(1)

        # ball drains and gets kicked to trough
        self.machine.switch_controller.process_switch('s_outhole', 1)
        self.advance_time_and_run(20)
        self.assertEqual(0, self.machine.ball_devices["bd_outhole"].balls)

        self.assertEqual(False, self.machine.game.tilted)

    @test_config("config_mechanical_eject.yaml")
    def test_mechanical_eject(self):
        """Test that tilt triggers auto launch."""
        self._add_tilt_handler()

        self._prepare_trough()
        self.start_game(2)

        self.assertBallsOnPlayfield(0)
        self.assertAvailableBallsOnPlayfield(1)
        self.assertEqual(1, self.machine.ball_devices["bd_launcher"].balls)

        self.assertFalse(self._is_tilted)
        self.machine.switch_controller.process_switch('s_tilt', 1)
        self.machine.switch_controller.process_switch('s_tilt', 0)
        self.advance_time_and_run(10)
        self.assertTrue(self._is_tilted)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(True, self.machine.game.tilted)

        # ball ejects
        self.assertBallsOnPlayfield(1)
        self.assertAvailableBallsOnPlayfield(1)
        self.assertEqual(0, self.machine.ball_devices["bd_launcher"].balls)

        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.advance_time_and_run(1)

        self.assertEqual(False, self.machine.game.tilted)

    def test_simple_tilt(self):
        self._add_tilt_handler()

        self._prepare_trough()
        self.start_game(2)

        # flipper actived
        self.assertTrue(self.machine.flippers["f_test"]._enabled)

        self.assertTrue(self.machine.mode_controller.is_active('tilt'))
        self.assertNotEqual(None, self.machine.game)

        # scoring should work
        self.post_event("test_scoring")
        self.assertPlayerVarEqual(100, "score")

        self.assertFalse(self._is_tilted)
        self.machine.switch_controller.process_switch('s_tilt', 1)
        self.machine.switch_controller.process_switch('s_tilt', 0)
        self.advance_time_and_run(1)
        self.assertTrue(self._is_tilted)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(True, self.machine.game.tilted)

        # flipper deactived
        self.assertFalse(self.machine.flippers["f_test"]._enabled)

        # scoring should no longer work
        self.assertPlayerVarEqual(100, "score")
        self.post_event("test_scoring")
        self.assertPlayerVarEqual(100, "score")

        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.advance_time_and_run(1)

        self.assertEqual(False, self.machine.game.tilted)

    def test_tilt_event(self):
        self._add_tilt_handler()

        self._prepare_trough()
        self.start_game(2)

        self.assertTrue(self.machine.mode_controller.is_active('tilt'))

        self.assertFalse(self._is_tilted)
        self.machine.events.post("tilt_event")
        self.advance_time_and_run(1)
        self.machine.events.post("tilt_event")
        self.advance_time_and_run(1)
        self.machine.events.post("tilt_event")
        self.advance_time_and_run(1)

        self.assertTrue(self._is_tilted)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(True, self.machine.game.tilted)

        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.advance_time_and_run(1)

        self.assertEqual(False, self.machine.game.tilted)

    def test_simple_tilt_ball_not_on_pf_yet(self):
        self._add_tilt_handler()

        self._prepare_trough()
        self.start_game(2)

        self.assertTrue(self.machine.mode_controller.is_active('tilt'))
        self.assertNotEqual(None, self.machine.game)

        self.assertFalse(self._is_tilted)
        self.machine.switch_controller.process_switch('s_tilt', 1)
        self.machine.switch_controller.process_switch('s_tilt', 0)
        self.advance_time_and_run(.1)
        self.assertTrue(self._is_tilted)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(True, self.machine.game.tilted)

        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.advance_time_and_run(1)

        self.assertEqual(False, self.machine.game.tilted)

    def test_tilt_warning(self):
        self._add_tilt_handler()

        self._prepare_trough()
        self.start_game(2)

        self.assertTrue(self.machine.mode_controller.is_active('tilt'))

        self.assertFalse(self._is_tilted)
        self.assertPlaceholderEvaluates(3, "mode.tilt.tilt_warnings_remaining")

        # multiple hits in 300ms window
        self.machine.switch_controller.process_switch('s_tilt_warning', 1)
        self.machine.switch_controller.process_switch('s_tilt_warning', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_tilt_warning', 1)
        self.machine.switch_controller.process_switch('s_tilt_warning', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_tilt_warning', 1)
        self.machine.switch_controller.process_switch('s_tilt_warning', 0)
        self.advance_time_and_run(1)
        self.assertFalse(self._is_tilted)
        self.assertNotEqual(None, self.machine.game)
        self.assertPlaceholderEvaluates(2, "mode.tilt.tilt_warnings_remaining")

        self.machine.switch_controller.process_switch('s_tilt_warning', 1)
        self.machine.switch_controller.process_switch('s_tilt_warning', 0)
        self.advance_time_and_run(1)
        self.assertFalse(self._is_tilted)
        self.assertNotEqual(None, self.machine.game)
        self.assertPlaceholderEvaluates(1, "mode.tilt.tilt_warnings_remaining")

        self.machine.switch_controller.process_switch('s_tilt_warning', 1)
        self.machine.switch_controller.process_switch('s_tilt_warning', 0)
        self.advance_time_and_run(1)
        self.assertTrue(self._is_tilted)
        self.assertPlaceholderEvaluates(0, "mode.tilt.tilt_warnings_remaining")
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(True, self.machine.game.tilted)

        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.advance_time_and_run(1)
        self.assertNotEqual(None, self.machine.game)
        self.assertPlaceholderEvaluates(3000.0, "mode.tilt.tilt_settle_ms_remaining")

        # wait for settle time (5s) since last s_tilt_warning hit
        self.advance_time_and_run(3.5)
        self.assertEqual(False, self.machine.game.tilted)

    def test_slam_tilt(self):
        self._add_tilt_handler()

        self._prepare_trough()

        self.assertEqual(None, self.machine.game)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(2, self.machine.ball_devices["bd_trough"].balls)
        self.machine.switch_controller.process_switch('s_start', 1)
        self.machine.switch_controller.process_switch('s_start', 0)
        self.advance_time_and_run(10)

        # flipper actived
        self.assertTrue(self.machine.flippers["f_test"]._enabled)

        self.assertTrue(self.machine.mode_controller.is_active('tilt'))
        self.assertNotEqual(None, self.machine.game)

        self.assertFalse(self._is_tilted)
        self.machine.switch_controller.process_switch('s_slam_tilt', 1)
        self.machine.switch_controller.process_switch('s_slam_tilt', 0)
        self.advance_time_and_run(1)
        self.assertNotEqual(None, self.machine.game)

        # flipper deactived
        self.assertFalse(self.machine.flippers["f_test"]._enabled)

        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.advance_time_and_run(1)

        self.assertEqual(None, self.machine.game)

        # test that it does not crash outside the game
        self.post_event("tilt_reset_warnings")
        self.advance_time_and_run()

    def test_carry_over_multiple_player(self):
        self._add_tilt_handler()
        self._prepare_trough()
        self.start_two_player_game()

        self.advance_time_and_run(10)
        self.assertBallsOnPlayfield(1)
        self.assertPlayerVarEqual(0, "tilt_warnings")

        # tilt machine
        self.hit_and_release_switch("s_tilt_warning")
        self.advance_time_and_run(2)
        self.assertFalse(self._is_tilted)
        self.assertPlayerVarEqual(1, "tilt_warnings")
        self.hit_and_release_switch("s_tilt_warning")
        self.advance_time_and_run(2)
        self.assertFalse(self._is_tilted)
        self.assertPlayerVarEqual(2, "tilt_warnings")

        self.hit_and_release_switch("s_tilt_warning")
        self.advance_time_and_run(2)
        self.assertTrue(self._is_tilted)
        self.assertPlayerVarEqual(3, "tilt_warnings")

        # tilt_warnings player var no longer increases because machine is tilted
        self.hit_and_release_switch("s_tilt_warning")
        self.advance_time_and_run(2)
        self.assertPlayerVarEqual(3, "tilt_warnings")
        self.hit_and_release_switch("s_tilt_warning")
        self.advance_time_and_run(2)
        self.assertPlayerVarEqual(3, "tilt_warnings")
        self.hit_and_release_switch("s_tilt_warning")
        self.advance_time_and_run(2)
        self.assertPlayerVarEqual(3, "tilt_warnings")
        self.assertPlayerNumber(1)

        self.drain_one_ball()
        self.advance_time_and_run(10)
        self.assertPlayerNumber(2)
        self.assertPlayerVarEqual(0, "tilt_warnings")
        self._is_tilted = False

        self.assertBallsOnPlayfield(1)

        self.hit_and_release_switch("s_tilt_warning")
        self.advance_time_and_run(2)
        self.assertPlayerVarEqual(1, "tilt_warnings")
        self.assertFalse(self._is_tilted)

    @test_config_directory("tests/machine_files/tilt_defaults/")
    def test_carry_over_single_player(self):
        self._add_tilt_handler()
        self._prepare_trough()
        self.start_game()

        self.advance_time_and_run(10)
        self.assertBallsOnPlayfield(1)
        self.assertPlayerVarEqual(0, "tilt_warnings")

        # tilt machine
        self.hit_and_release_switch("s_tilt_warning")
        self.advance_time_and_run(2)
        self.assertFalse(self._is_tilted)
        self.assertPlayerVarEqual(1, "tilt_warnings")
        self.hit_and_release_switch("s_tilt_warning")
        self.advance_time_and_run(2)
        self.assertFalse(self._is_tilted)
        self.assertPlayerVarEqual(2, "tilt_warnings")

        self.hit_and_release_switch("s_tilt_warning")
        self.advance_time_and_run(2)
        self.assertTrue(self._is_tilted)
        self.assertPlayerVarEqual(0, "tilt_warnings")

        # tilt_warnings player var no longer increases because machine is tilted
        self.hit_and_release_switch("s_tilt_warning")
        self.advance_time_and_run(2)
        self.assertPlayerVarEqual(0, "tilt_warnings")

        self.hit_and_release_switch("s_tilt_warning")
        self.advance_time_and_run(2)
        self.assertPlayerVarEqual(0, "tilt_warnings")
        self.hit_and_release_switch("s_tilt_warning")
        self.advance_time_and_run(2)
        self.assertPlayerVarEqual(0, "tilt_warnings")
        self.assertPlayerNumber(1)

        # advance to ball 2
        self.drain_one_ball()
        self.advance_time_and_run(10)
        self.assertPlayerNumber(1)

        self._is_tilted = False

        self.assertBallNumber(2)
        self.assertBallsOnPlayfield(1)
        self.assertPlayerVarEqual(0, "tilt_warnings")

        # generate tilt warning and ensure no carry over from previous ball
        self.hit_and_release_switch("s_tilt_warning")
        self.advance_time_and_run(2)
        self.assertPlayerVarEqual(1, "tilt_warnings")
        self.assertFalse(self._is_tilted)



"""Contains show related classes."""
import re
from collections import namedtuple

from typing import List, Dict, Any, Optional

from mpf.core.assets import AssetPool
from mpf.core.config_validator import RuntimeToken
from mpf.core.utility_functions import Util
from mpf.exceptions.config_file_error import ConfigFileError

MYPY = False
if MYPY:   # pragma: no cover
    from typing import NoReturn     # pylint: disable-msg=cyclic-import,unused-import

__api__ = ['Show', 'RunningShow', 'ShowPool']

ShowConfig = namedtuple("ShowConfig", ["name", "priority", "speed", "loops", "sync_ms", "manual_advance", "show_tokens",
                                       "events_when_played", "events_when_stopped", "events_when_looped",
                                       "events_when_paused", "events_when_resumed", "events_when_advanced",
                                       "events_when_stepped_back", "events_when_updated", "events_when_completed"])


class ShowPool(AssetPool):

    """A pool of shows."""

    __slots__ = []  # type: List[str]

    def __repr__(self):
        """Return str representation."""
        return '<ShowPool: {}>'.format(self.name)

    # pylint: disable-msg=too-many-arguments
    def play_with_config(self, show_config: ShowConfig, start_time=None, start_callback=None, stop_callback=None,
                         start_step=None) -> "RunningShow":
        """Play asset from pool with config."""
        return self.asset.play_with_config(show_config, start_time, start_callback, stop_callback, start_step)

    # pylint: disable-msg=too-many-arguments
    # pylint: disable-msg=too-many-locals
    def play(self, priority=0, speed=1.0, start_step=1, callback=None,
             loops=-1, sync_ms=None, manual_advance=False, show_tokens=None,
             events_when_played=None, events_when_stopped=None,
             events_when_looped=None, events_when_paused=None,
             events_when_resumed=None, events_when_advanced=None,
             events_when_stepped_back=None, events_when_updated=None,
             events_when_completed=None, start_time=None, start_callback=None) -> "RunningShow":
        """Play asset from pool."""
        return self.asset.play(priority, speed, start_step, callback,
                               loops, sync_ms, manual_advance, show_tokens,
                               events_when_played, events_when_stopped,
                               events_when_looped, events_when_paused,
                               events_when_resumed, events_when_advanced,
                               events_when_stepped_back, events_when_updated,
                               events_when_completed, start_time, start_callback)


# pylint: disable-msg=too-many-instance-attributes
class Show:

    """A show which can be instantiated."""

    attribute = 'shows'
    path_string = 'shows'
    config_section = 'shows'
    disk_asset_section = 'file_shows'
    extensions = tuple('yaml')
    class_priority = 100
    pool_config_section = 'show_pools'
    asset_group_class = ShowPool

    __slots__ = ["_autoplay_settings", "tokens", "token_values", "token_keys", "name", "total_steps", "show_steps",
                 "_step_cache", "machine"]

    def __init__(self, machine, name):
        """Initialise show."""
        self.machine = machine
        self._autoplay_settings = dict()
        self.tokens = set()
        self.token_values = dict()
        self.token_keys = dict()

        self.name = name
        self.total_steps = None
        self.show_steps = []      # type: List[Dict[str, Any]]
        self._step_cache = {}

    def __lt__(self, other):
        """Compare two instances."""
        return id(self) < id(other)

    def _get_duration(self, data, step_num, total_step_time):
        total_steps_num = len(data)
        step = data[step_num]
        if 'duration' not in step:
            if step_num == total_steps_num - 1:
                # special case with an empty last step (but longer than 1 step)
                if 'time' in step and len(step) == 1 and step_num != 0:
                    return False
                return 1
            if 'time' in data[step_num + 1]:
                next_step_time = data[step_num + 1]['time']
                if str(next_step_time)[0] == "+":
                    return Util.string_to_secs(next_step_time)

                if total_step_time < 0:     # pragma: no cover
                    self._show_validation_error("Absolute timing in step {} not possible because "
                                                "there was a duration of -1 before".format(step_num), 5)
                return Util.string_to_secs(next_step_time) - total_step_time

            return 1

        if step_num < total_steps_num - 1 and 'time' in data[step_num + 1]:     # pragma: no cover
            self._show_validation_error("Found invalid 'time' entry in step after {} which contains a duration. "
                                        "Remove either of them!".format(step_num), 2)
        return Util.string_to_secs(step['duration'])

    def load(self, data: Optional[Dict]):
        """Load show configuration."""
        self.show_steps = list()

        if not isinstance(data, list):    # pragma: no cover
            self._show_validation_error("Show {} does not appear to be a valid show "
                                        "config. It should be a list of steps. Did you forget the hyphen at the start "
                                        "of your step?".format(self.name), 1)

        if not data:    # pragma: no cover
            self._show_validation_error("Cannot load empty show", 6)

        total_step_time = 0

        # add empty first step if show does not start right away
        if 'time' in data[0] and data[0]['time'] != 0:
            self.show_steps.append({'duration': Util.string_to_secs(data[0]['time'])})
            total_step_time = Util.string_to_secs(data[0]['time'])

        # Loop over all steps in the show file
        for step_num, step in enumerate(data):
            actions = dict()

            # Note: all times are stored/calculated in seconds.

            # Step time can be specified as either an absolute time elapsed
            # (from the beginning of the show) or a relative time (time elapsed
            # since the previous step).  Time strings starting with a plus sign
            # (+) are treated as relative times.

            # Step times are all converted to relative times internally (time
            # since the previous step).

            # Make sure there is a time entry for each step in the show file.
            duration = self._get_duration(data, step_num, total_step_time)

            # special case: empty last step
            if duration is False:
                break
            if duration == 0:     # pragma: no cover
                self._show_validation_error("Step {} has 0 duration".format(step_num), 7)

            # Calculate the time since previous step
            actions['duration'] = duration

            if duration > 0 and total_step_time >= 0:
                total_step_time += duration
            else:
                total_step_time = -1

            # Now process show step actions
            self._process_step_actions(step, actions)

            self.show_steps.append(actions)

        # Count how many total steps are in the show. We need this later
        # so we can know when we're at the end of a show
        self.total_steps = len(self.show_steps)
        if self.total_steps == 0:   # pragma: no cover
            self._show_validation_error('Show "{}" is empty', 2)

        self._get_tokens()

    def _show_validation_error(self, msg, error_code) -> "NoReturn":  # pragma: no cover
        raise ConfigFileError('"{}" >> {}'.format(self.name, msg), error_code, "show", self.name)

    def _process_step_actions(self, step, actions):
        if not isinstance(step, dict):
            raise AssertionError('Steps in show "{}" need to be dicts.'.format(self.name))
        for key, value in step.items():

            # key: the section of the show, like 'leds'
            # value: dic of express settings or dic of dics w full settings

            # check to see if we know how to process this kind of entry
            if key in self.machine.show_controller.show_players.keys():
                try:
                    actions[key] = \
                        self.machine.show_controller.show_players[key].validate_config_entry(value, self.name)
                # If something in the show triggered a config error, bubble it up to preserve logger and context
                except ConfigFileError as e:
                    e.extend('Show "{}"'.format(self.name))
                    raise e

            elif key not in ('duration', 'time'):   # pragma: no cover
                for player in self.machine.show_controller.show_players.values():
                    if key == player.config_file_section or key == player.machine_collection_name or \
                            key + "s" == player.show_section:
                        self._show_validation_error(
                            'Invalid section "{}:" found. Did you mean "{}:" instead?'.format(
                                key, player.show_section), 3)
                self._show_validation_error('Invalid section "{}:" found.'.format(key), 4)

    def _get_tokens(self):
        self._walk_show(self.show_steps)

    def _walk_show(self, data, path=None, list_index=None):
        # walks a list of dicts, checking tokens
        if not path:
            path = list()

        if isinstance(data, dict):
            for k, v in data.items():
                self._check_token(path, k, 'key')
                self._walk_show(v, path + [k])

        elif isinstance(data, list):
            for i in data:
                self._check_token(path, i, 'key')
                if list_index is None:
                    list_index = 0
                else:
                    list_index += 1
                self._walk_show(i, path + [list_index], list_index)

        else:
            self._check_token(path, data, 'value')

    @classmethod
    def _copy_recursive(cls, data):
        if isinstance(data, dict):
            new_dict = dict()
            for k, v in data.items():
                new_dict[k] = cls._copy_recursive(v)
            return new_dict
        if isinstance(data, list):
            new_list = list()
            for i in data:
                new_list.append(cls._copy_recursive(i))
            return new_list
        return data

    def get_show_steps(self):
        """Return a copy of the show steps."""
        copied_steps = []
        for step in self.show_steps:
            copied_steps.append(self._copy_recursive(step))
        return copied_steps

    def _check_token(self, path, data, token_type):
        if isinstance(data, RuntimeToken):
            self._add_token(data, data.token, path, token_type)
            return
        if not isinstance(data, str):
            return

        results = re.findall(r"\(([^)]+)\)", data)
        if results:
            for result in results:
                self._add_token(data, result, path, token_type)

    def _add_token(self, placeholder, token, path, token_type):

        if token not in self.tokens:
            self.tokens.add(token)

        if token_type == 'key':
            if token not in self.token_keys:
                self.token_keys[token] = list()
            self.token_keys[token].append(path + [placeholder])

        elif token_type == 'value':
            if token not in self.token_values:
                self.token_values[token] = list()
            self.token_values[token].append(path)

    # pylint: disable-msg=too-many-arguments
    def play_with_config(self, show_config: ShowConfig, start_time=None, start_running=True,
                         start_callback=None, stop_callback=None, start_step=None) -> "RunningShow":
        """Play this show with config."""
        if not start_time:
            start_time = self.machine.clock.get_time()
        running_show = RunningShow(machine=self.machine,
                                   show=self,
                                   start_time=start_time,
                                   start_step=int(start_step),
                                   start_running=start_running,
                                   callback=stop_callback,
                                   start_callback=start_callback,
                                   show_config=show_config)

        return running_show

    # pylint: disable-msg=too-many-arguments
    # pylint: disable-msg=too-many-locals
    def play(self, priority=0, speed=1.0, start_step=1, callback=None,
             loops=-1, sync_ms=None, manual_advance=False, show_tokens=None,
             events_when_played=None, events_when_stopped=None,
             events_when_looped=None, events_when_paused=None,
             events_when_resumed=None, events_when_advanced=None,
             events_when_stepped_back=None, events_when_updated=None,
             events_when_completed=None, start_time=None, start_callback=None,
             start_running=True) -> "RunningShow":
        """Play a Show.

        There are many parameters you can use here which
        affect how the show is played. This includes things like the playback
        speed, priority, etc. These are
        all set when the show plays. (For example, you could have a Show
        file which lights a bunch of lights sequentially in a circle pattern,
        but you can have that circle "spin" as fast as you want depending on
        how you play the show.)

        Args:
        ----
            priority: Integer value of the relative priority of this show. If
                there's ever a situation where multiple shows want to control
                the same item, the one with the higher priority will win.
                ("Higher" means a bigger number, so a show with priority 2 will
                override a priority 1.)
            speed: Float of how fast your show runs. Your Show files
                specify step times in actual time values.  When you play a
                show,
                you specify a playback rate factor that is applied to the time
                values in the show (divides the relative show times). The
                default value is 1.0 (uses the actual time values in specified
                in the show), but you might want to speed up (speed
                values > 1.0) or slow down (speed values < 1.0) the
                playback rate.  If you want your show to play twice as fast
                (finish in half the time), you want all your time values to be
                half of the specified values in the show so you would use a
                speed value of 2.0.  To make the show take twice as
                long
                to finish, you would a speed value of 0.5.
            start_step: Integer of which step in the show file the show
                should start in. Usually this is 1 (start at the beginning
                of the show) but it's nice to start part way through. Also
                used for restarting shows that you paused. A negative value
                will count backwards from the end (-1 is the last position,
                -2 is second to last, etc.). Note this is the "human readable"
                step, so the first step is 1, not 0.
            start_running: Boolean of whether this show should start in a running
                state, i.e. begin advancing through steps. If false, the show will
                load the first step and enter a paused state. Default value is true.
            callback: A callback function that is invoked when the show is
                stopped.
            loops: Integer of how many times you want this show to repeat
                before stopping. A value of -1 means that it repeats
                indefinitely. If the show only has one step, loops will be set
                to 0, regardless of the actual number of loops
            sync_ms: Number of ms of the show sync cycle. A value of zero means
                this show will also start playing immediately. A value of None
                means the mpf:default_show_sync_ms will be used.
            manual_advance: Boolean that controls whether this show should be
                advanced manually (e.g. time values are ignored and the show
                doesn't move to the next step until it's told to.) Default is
                False.
            show_tokens: Replacement tokens for the show
            events_when_played: Events to post when show is started
            events_when_stopped: Events to post when show is stopped
            events_when_looped: Events to post when show looped
            events_when_paused: Events to post when show is paused
            events_when_resumed: Events to post when show is resumed after it has been
            events_when_advanced: Events to post when show is advanced
            events_when_stepped_back: Events to post when show is stepped back
            events_when_updated: Events to post when show is updated
            events_when_completed: Events to post when show completed
            start_time: Time when this show was started. This used to synchronize shows
            start_callback: Callback when the show is first started

        Return the RunningShow() instance if this show plays now, or False if
        the show is not loaded. (In this case the show will be loaded and will
        automatically play once its loaded.)
        """
        if not show_tokens:
            show_tokens = dict()

        # todo if we want to enforce that show_tokens match the tokens in the
        # show exactly, uncomment below and remove the following if.
        # however we don't do this today because of the default 'off' show
        # that's used since it has lights and leds, so we'll have to think
        # about this.

        # if set(show_tokens.keys()) != self.tokens:
        #     raise ValueError('Token mismatch while playing show "{}". Tokens '
        #                      'expected: {}. Tokens submitted: {}'.format(
        #                      self.name, self.tokens, set(show_tokens.keys())))

        if not set(show_tokens.keys()).issubset(self.tokens):   # pragma: no cover
            raise ValueError('Token mismatch while playing show "{}". Tokens '
                             'expected: {}. Tokens submitted: {}'.
                             format(self.name, self.tokens, set(show_tokens.keys())))

        show_config = self.machine.show_controller.create_show_config(
            self.name, priority, speed, loops, sync_ms, manual_advance, show_tokens, events_when_played,
            events_when_stopped, events_when_looped, events_when_paused, events_when_resumed, events_when_advanced,
            events_when_stepped_back, events_when_updated, events_when_completed)

        return self.play_with_config(show_config, start_time, start_running, start_callback, callback, start_step)

    def get_show_steps_with_token(self, show_tokens):
        """Get show steps and replace additional tokens."""
        if show_tokens and self.tokens:
            token_hash = hash(str(show_tokens))
            if token_hash in self._step_cache:
                return self._step_cache[token_hash]

            show_steps = self.get_show_steps()
            # if we need to replace more tokens copy the show
            self._replace_token_values(show_steps, show_tokens)
            self._replace_token_keys(show_steps, show_tokens)

            for step in show_steps:
                for key, value in step.items():
                    if key in self.machine.show_controller.show_players.keys():
                        step[key] = self.machine.show_controller.show_players[key].expand_config_entry(value)

            self._step_cache[token_hash] = show_steps
            return show_steps

        # otherwise return show steps. the caller should not change them
        return self.show_steps

    def _replace_token_values(self, show_steps, show_tokens):
        for token, replacement in show_tokens.items():
            if token in self.token_values:
                for token_path in self.token_values[token]:
                    target = show_steps
                    for x in token_path[:-1]:
                        target = target[x]

                    if isinstance(target[token_path[-1]], RuntimeToken):
                        target[token_path[-1]] = target[token_path[-1]].validator_function(replacement, None)
                    elif target[token_path[-1]] == "(" + token + ")":
                        target[token_path[-1]] = replacement
                    else:
                        target[token_path[-1]] = target[token_path[-1]].replace("(" + token + ")", replacement)
        return show_steps

    def _replace_token_keys(self, show_steps, show_tokens):
        keys_replaced = dict()
        # pylint: disable-msg=too-many-nested-blocks
        for token, replacement in show_tokens.items():
            if token in self.token_keys:
                key_name = '({})'.format(token)
                for token_path in self.token_keys[token]:
                    target = show_steps
                    token_str = ""
                    for x in token_path[:-1]:
                        if token_str in keys_replaced:
                            x = keys_replaced[token_str + str(x) + "-"]
                        token_str += str(x) + "-"

                        target = target[x]
                    use_string_replace = bool(token_path[-1] != "(" + token + ")")

                    final_key = token_path[-1]
                    # check if key has been replaced before
                    final_key = keys_replaced.get(final_key, final_key)

                    if use_string_replace:
                        replaced_key = final_key.replace("(" + token + ")", replacement)
                    else:
                        replaced_key = replacement

                    if final_key in target:
                        target[replaced_key] = target.pop(final_key)
                    else:
                        raise KeyError("Could not find token {} ({}) in {}".format(final_key, key_name, target))

                    keys_replaced[token_str] = replaced_key
        return show_steps


# This class is more or less a container
# pylint: disable-msg=too-many-instance-attributes
class RunningShow:

    """A running instance of a show."""

    __slots__ = ["machine", "show", "show_steps", "show_config", "callback", "start_step", "start_running",
                 "start_callback", "_delay_handler", "next_step_index", "current_step_index", "next_step_time",
                 "name", "loops", "id", "_players", "debug", "_stopped", "_total_steps", "context"]

    # pylint: disable-msg=too-many-arguments
    # pylint: disable-msg=too-many-locals
    def __init__(self, machine, show, start_step, start_running, callback, start_time, start_callback, show_config):
        """Initialise an instance of a show."""
        self.machine = machine
        self.show = show
        self.show_config = show_config
        self.callback = callback
        self.start_step = start_step
        self.start_running = start_running
        self.start_callback = start_callback
        self._delay_handler = None
        self.next_step_index = None
        self.current_step_index = None
        self.next_step_time = start_time
        self.name = show.name
        self.loops = self.show_config.loops

        self.id = self.machine.show_controller.get_next_show_id()
        self.context = "show_{}".format(self.id)
        self._players = set()

        self.debug = False
        self._stopped = False
        self._total_steps = None
        self.show_steps = self.show.get_show_steps_with_token(self.show_config.show_tokens)
        self._start_play()

    def _start_play(self):
        if self._stopped:
            return

        self._total_steps = len(self.show_steps)

        if self.start_step > 0:
            self.next_step_index = self.start_step - 1
        elif self.start_step < 0:
            self.next_step_index = self.start_step % self._total_steps
        else:
            self.next_step_index = 0

        # Figure out the show start time
        if self.show_config.sync_ms:
            # calculate next step based on synchronized start time
            self.next_step_time += (self.show_config.sync_ms / 1000.0) - (self.next_step_time %
                                                                          (self.show_config.sync_ms / 1000.0))
            # but wait relative to real time
            delay_secs = self.next_step_time - self.machine.clock.get_time()
            self._delay_handler = self.machine.clock.schedule_once(
                self._start_now, delay_secs)
        else:  # run now
            self._start_now()

    def _post_events(self, events):
        for event in events:
            self.machine.events.post(event)

    def __repr__(self):
        """Return str representation."""
        return 'Running Show Instance: "{}" {} {}'.format(self.name, self.show_config.show_tokens, self.next_step_index)

    @property
    def stopped(self):
        """Return if stopped."""
        return self._stopped

    def stop(self):
        """Stop show."""
        if self._stopped:
            return

        self._stopped = True

        # if the start callback has never been called then call it now
        if self.start_callback:
            self.start_callback()
            self.start_callback = None

        self._remove_delay_handler()

        # clear context in used players
        for player in self._players:
            self.machine.show_controller.show_players[player].show_stop_callback(self.context)

        self._players = set()

        if self.callback and callable(self.callback):
            self.callback()

        if self.show_config.events_when_stopped:
            self._post_events(self.show_config.events_when_stopped)

    def _remove_delay_handler(self):
        if self._delay_handler:
            self.machine.clock.unschedule(self._delay_handler)
            self._delay_handler = None

    def pause(self):
        """Pause show."""
        self._remove_delay_handler()
        if self.show_config.events_when_paused:
            self._post_events(self.show_config.events_when_paused)

    def resume(self):
        """Resume paused show."""
        self.next_step_time = self.machine.clock.get_time()
        self._run_next_step(post_events=self.show_config.events_when_resumed)

    def update(self, **kwargs):
        """Update show.

        Updates the values of a show while it runs. Currently supports only speed
        and manual_advance properties.
        """
        updated_values = {k: v for k, v in kwargs.items() if v is not None and v != getattr(self.show_config, k)}
        if updated_values:
            self.show_config = self.show_config._replace(**updated_values)

        self._post_events(['updated'])

    def advance(self, steps=1, show_step=None):
        """Manually advance this show to the next step."""
        self._remove_delay_handler()

        if steps != 1:
            self.next_step_index += steps - 1
        elif show_step is not None:
            if not isinstance(show_step, int) or show_step < 0:
                raise AssertionError('Cannot advance {} to step "{}" as that is'
                                     'not a valid step number.'.format(self, show_step))
            self.next_step_index = show_step - 1

        self._run_next_step(post_events=self.show_config.events_when_advanced)

    def step_back(self, steps=1):
        """Manually step back this show to a previous step."""
        self._remove_delay_handler()

        self.next_step_index -= steps + 1

        self._run_next_step(post_events=self.show_config.events_when_stepped_back)

    def _start_now(self) -> None:
        """Start playing the show."""
        if self.start_callback:
            self.start_callback()
            self.start_callback = None
        pause_after_step = not self.start_running
        self._run_next_step(post_events=self.show_config.events_when_played,
                            pause_after_step=pause_after_step)

    def _run_next_step(self, post_events=None, pause_after_step=False) -> None:
        """Run the next show step."""
        events = []
        if post_events:
            events.extend(post_events)

        if self.next_step_index < 0:
            self.next_step_index %= self._total_steps

        # if we're at the end of the show
        if self.next_step_index >= self._total_steps:

            if self.loops > 0:
                self.loops -= 1
                self.next_step_index = 0
                if self.show_config.events_when_looped:
                    events.extend(self.show_config.events_when_looped)
            elif self.loops < 0:
                self.next_step_index = 0
                if self.show_config.events_when_looped:
                    events.extend(self.show_config.events_when_looped)
            else:
                self.stop()
                if self.show_config.events_when_completed:
                    events.extend(self.show_config.events_when_completed)
                self._post_events(events)
                return

        self.current_step_index = self.next_step_index

        for item_type, item_dict in self.show_steps[self.current_step_index].items():

            if item_type == 'duration':
                continue

            try:
                player = self.machine.show_controller.show_players[item_type]
            except KeyError:
                raise ValueError("Invalid entry in show: {}".format(item_type))

            player.show_play_callback(
                settings=item_dict,
                context=self.context,
                calling_context=self.current_step_index,
                priority=self.show_config.priority,
                show_tokens=self.show_config.show_tokens,
                start_time=self.next_step_time)

            self._players.add(item_type)

        if events:
            self._post_events(events)

        self.next_step_index += 1

        time_to_next_step = self.show_steps[self.current_step_index]['duration'] / self.show_config.speed
        if not self.show_config.manual_advance and time_to_next_step > 0 and not pause_after_step:
            self.next_step_time += time_to_next_step
            self._delay_handler = self.machine.clock.loop.call_at(when=self.next_step_time,
                                                                  callback=self._run_next_step)



"""Test queue event player."""
from mpf.tests.MpfTestCase import MpfTestCase


class TestQueueEventPlayer(MpfTestCase):

    def get_config_file(self):
        return 'test_queue_event_player.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/event_players/'

    def _queue1(self, queue, **kwargs):
        del kwargs
        queue.wait()
        self.queue1 = queue

    def _queue2(self, queue, **kwargs):
        del kwargs
        queue.wait()
        self.queue2 = queue

    def test_queue_event_player(self):
        self.mock_event("queue_event1_finished")
        self.queue1 = None
        self.queue2 = None
        self.machine.events.add_handler("queue_event1", self._queue1, priority=2)
        self.machine.events.add_handler("queue_event1", self._queue2, priority=1)
        self.post_event("play")
        self.advance_time_and_run()

        self.assertIsNotNone(self.queue1)
        self.assertIsNone(self.queue2)

        self.queue1.clear()
        self.advance_time_and_run()
        self.assertIsNotNone(self.queue2)
        self.assertEventNotCalled("queue_event1_finished")

        self.queue2.clear()
        self.advance_time_and_run()
        self.assertEventCalled("queue_event1_finished")

    def _cb(self, **kwargs):
        del kwargs
        self._done = True

    def test_queue_relay_player(self):
        self._done = False
        self.mock_event("relay_start")
        self.mock_event("relay2_start")

        # post queue event
        self.machine.events.post_queue("relay", callback=self._cb, something=7)
        self.advance_time_and_run()

        # should run the first relay only
        self.assertFalse(self._done)
        self.assertEventCalledWith("relay_start")
        self.assertEventNotCalled("relay2_start")

        # first relay done. should trigger the second
        self.post_event("relay_done")
        self.advance_time_and_run()
        self.assertEventCalled("relay2_start")
        self.assertFalse(self._done)

        # second done. should trigger cb
        self.post_event("relay2_done")
        self.advance_time_and_run()
        self.assertTrue(self._done)

    def test_queue_relay_player_with_pass_args(self):
        self._done = False
        self.mock_event("relay_with_args_start")

        # post queue event
        self.machine.events.post_queue("relay_with_args", callback=self._cb, test=42)
        self.advance_time_and_run()

        # should run the first relay only
        self.assertFalse(self._done)
        self.assertEventCalledWith("relay_with_args_start", test=42)

        # relay done
        self.post_event("relay_with_args_done")
        self.advance_time_and_run()
        self.assertTrue(self._done)

    def test_broken_queue_relay_player(self):
        with self.assertRaises(Exception) as e:
            self.post_event("relay")
        self.assertIsInstance(e.exception.__cause__, AssertionError)
        self.assertEqual(str(e.exception.__cause__), "Can only use queue relay player with queue event.")

        # reset exception
        self._exception = None

    def test_queue_relay_player_in_mode(self):
        self._done = False
        self.mock_event("relay3_start")

        self.machine.modes["mode1"].start()

        # post queue event
        self.machine.events.post_queue("relay3", callback=self._cb)
        self.advance_time_and_run()

        # should run the relay
        self.assertFalse(self._done)
        self.assertEventCalled("relay3_start")

        # relay done. should trigger cb
        self.post_event("relay3_done")
        self.advance_time_and_run()
        self.assertTrue(self._done)

        # stop and start mode again
        self.machine.modes["mode1"].stop()
        self.advance_time_and_run()

        self._done = False
        self.mock_event("relay3_start")
        self.machine.modes["mode1"].start()

        # post queue event
        self.machine.events.post_queue("relay3", callback=self._cb)
        self.advance_time_and_run()

        # should run the relay
        self.assertFalse(self._done)
        self.assertEventCalled("relay3_start")

        # relay done. should trigger cb
        self.post_event("relay3_done")
        self.advance_time_and_run()
        self.assertTrue(self._done)

        # post queue event again
        self._done = False
        self.machine.events.post_queue("relay3", callback=self._cb)
        self.advance_time_and_run()
        self.assertFalse(self._done)

        # stop mode
        self.machine.modes["mode1"].stop()
        self.advance_time_and_run()

        # should trigger cb without waiting for event
        self.assertTrue(self._done)

        # relay should not crash it
        self.post_event("relay3_done")
        self.advance_time_and_run()



"""Contains the base class for autofire coil devices."""
from mpf.core.delays import DelayManager
from mpf.core.device_monitor import DeviceMonitor
from mpf.core.events import event_handler
from mpf.core.platform_controller import SwitchRuleSettings, DriverRuleSettings, PulseRuleSettings, HardwareRule

from mpf.core.system_wide_device import SystemWideDevice

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import
    from typing import List, Optional  # pylint: disable-msg=cyclic-import,unused-import


@DeviceMonitor(_enabled="enabled")
class AutofireCoil(SystemWideDevice):

    """Autofire coils which fire based on switch hits with a hardware rule.

    Coils in the pinball machine which should fire automatically based on
    switch hits using defined hardware switch rules.

    Autofire coils work with rules written to the hardware pinball controller
    that allow them to respond "instantly" to switch hits versus waiting for
    the lag of USB and the host computer.

    Examples of Autofire Coils are pop bumpers, slingshots, and kicking
    targets. (Flippers use the same autofire rules under the hood, but flipper
    devices have their own device type in MPF.

    """

    config_section = 'autofire_coils'
    collection = 'autofires'
    class_label = 'autofire'

    __slots__ = ["_enabled", "_rule", "delay", "_ball_search_in_progress", "_timeout_watch_time", "_timeout_max_hits",
                 "_timeout_disable_time", "_timeout_hits"]

    def __init__(self, machine: "MachineController", name: str) -> None:
        """Initialise autofire."""
        self._enabled = False
        self._rule = None       # type: Optional[HardwareRule]
        super().__init__(machine, name)
        self.delay = DelayManager(self.machine)
        self._ball_search_in_progress = False
        self._timeout_watch_time = None
        self._timeout_max_hits = None
        self._timeout_disable_time = None
        self._timeout_hits = []     # type: List[float]

    async def _initialize(self):
        await super()._initialize()
        if self.config['ball_search_order']:
            self.config['playfield'].ball_search.register(
                self.config['ball_search_order'], self._ball_search, self.name)
        # pulse is handled via rule but add a handler so that we take notice anyway
        self.config['switch'].add_handler(self._hit)
        if self.config['timeout_watch_time']:
            self._timeout_watch_time = self.config['timeout_watch_time'] / 1000
            self._timeout_max_hits = self.config['timeout_max_hits']
            self._timeout_disable_time = self.config['timeout_disable_time']

        if '{}_active'.format(self.config['playfield'].name) in self.config['switch'].tags:
            self.raise_config_error(
                "Autofire device '{}' uses switch '{}' which has a "
                "'{}_active' tag. This is handled internally by the device. Remove the "
                "redundant '{}_active' tag from that switch.".format(
                    self.name, self.config['switch'].name, self.config['playfield'].name,
                    self.config['playfield'].name), 1)

    @event_handler(1)
    def event_enable(self, **kwargs):
        """Handle enable control event.

        To prevent multiple rules at the same time we prioritize disable > enable.
        """
        del kwargs
        self.enable()

    def enable(self):
        """Enable the autofire device.

        This causes the coil to respond to the switch hits. This is typically
        called when a ball starts to enable the slingshots, pops, etc.

        Note that there are several options for both the coil and the switch
        which can be incorporated into this rule, including recycle times,
        switch debounce, reversing the switch (fire the coil when the switch
        goes inactive), etc. These rules vary by hardware platform. See the
        user documentation for the hardware platform for details.

        Args:
        ----
            **kwargs: Not used, just included so this method can be used as an
                event callback.

        """
        if self._enabled:
            return
        self._enabled = True

        self.debug_log("Enabling")

        if self.config['coil_overwrite'].get('recycle', None) is not None:
            # if coil_overwrite is set use it
            recycle = self.config['coil_overwrite']['recycle']
        else:
            # otherwise load the default from the coil and turn None to True
            recycle = self.config['coil'].config['default_recycle'] in (True, None)

        if self.config['switch_overwrite'].get('debounce', None) is not None:
            # if switch_overwrite is set use it
            debounce = self.config['switch_overwrite']['debounce'] == "normal"
        else:
            # otherwise load the default from the switch and turn auto into False
            debounce = self.config['switch'].config['debounce'] == "normal"

        if not self.config['coil_pulse_delay']:
            self._rule = self.machine.platform_controller.set_pulse_on_hit_rule(
                SwitchRuleSettings(switch=self.config['switch'], debounce=debounce,
                                   invert=self.config['reverse_switch']),
                DriverRuleSettings(driver=self.config['coil'], recycle=recycle),
                PulseRuleSettings(duration=self.config['coil_overwrite'].get('pulse_ms', None),
                                  power=self.config['coil_overwrite'].get('pulse_power', None))
            )
        else:
            self._rule = self.machine.platform_controller.set_delayed_pulse_on_hit_rule(
                SwitchRuleSettings(switch=self.config['switch'], debounce=debounce,
                                   invert=self.config['reverse_switch']),
                DriverRuleSettings(driver=self.config['coil'], recycle=recycle),
                self.config['coil_pulse_delay'],
                PulseRuleSettings(duration=self.config['coil_overwrite'].get('pulse_ms', None),
                                  power=self.config['coil_overwrite'].get('pulse_power', None))
            )

    @event_handler(10)
    def event_disable(self, **kwargs):
        """Handle disable control event.

        To prevent multiple rules at the same time we prioritize disable > enable.
        """
        del kwargs
        self.disable()

    def disable(self):
        """Disable the autofire device.

        This is typically called at the end of a ball and when a tilt event
        happens.

        Args:
        ----
            **kwargs: Not used, just included so this method can be used as an
                event callback.

        """
        self.delay.remove("_timeout_enable_delay")

        if not self._enabled:
            return
        self._enabled = False

        self.debug_log("Disabling")
        self.machine.platform_controller.clear_hw_rule(self._rule)

    def _hit(self):
        """Rule was triggered."""
        if not self._enabled:
            return
        if not self._ball_search_in_progress:
            self.config['playfield'].mark_playfield_active_from_device_action()
        if self._timeout_watch_time:
            current_time = self.machine.clock.get_time()
            self._timeout_hits = [t for t in self._timeout_hits if t > current_time - self._timeout_watch_time / 1000.0]
            self._timeout_hits.append(current_time)

            if len(self._timeout_hits) >= self._timeout_max_hits:
                self.disable()
                self.delay.add(self._timeout_disable_time, self.enable, "_timeout_enable_delay")

    def _ball_search(self, phase, iteration):
        del phase
        del iteration
        self.delay.reset(ms=200, callback=self._ball_search_ignore_done, name="ball_search_ignore_done")
        self._ball_search_in_progress = True
        self.config['coil'].pulse()
        return True

    def _ball_search_ignore_done(self):
        """We no longer expect any fake hits."""
        self._ball_search_in_progress = False



"""Interface for a hardware sound platform."""
from typing import List

import abc


class HardwareSoundPlatformInterface(metaclass=abc.ABCMeta):

    """Interface for a hardware sound platform."""

    __slots__ = []  # type: List[str]

    @abc.abstractmethod
    def play_sound(self, number: int, track: int = 1):
        """Play a sound."""
        raise NotImplementedError

    @abc.abstractmethod
    def play_sound_file(self, file: str, platform_options: dict, track: int = 1):
        """Play a sound file."""
        raise NotImplementedError

    @abc.abstractmethod
    def text_to_speech(self, text: str, platform_options: dict, track: int = 1):
        """Text to speech output."""
        raise NotImplementedError

    @abc.abstractmethod
    def set_volume(self, volume: float, track: int = 1):
        """Set volume."""
        raise NotImplementedError

    @abc.abstractmethod
    def stop_all_sounds(self, track: int = 1):
        """Play a sound."""
        raise NotImplementedError



from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase


class TestStateMachine(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/state_machine/'

    def test_state_machine(self):
        placeholder_str = "device.state_machines.my_state.state"
        placeholder = self.machine.placeholder_manager.build_raw_template(placeholder_str)
        placeholder_value, placeholder_future = placeholder.evaluate_and_subscribe([])
        self.assertPlaceholderEvaluates("start", placeholder_str)
        self.assertEqual("start", self.machine.state_machines["my_state"].state)

        self.mock_event("step1_start")
        self.mock_event("step1_stop")
        self.mock_event("going_to_step2")
        self.post_event("state_machine_proceed")
        self.assertPlaceholderEvaluates("step1", placeholder_str)
        self.assertEqual("step1", self.machine.state_machines["my_state"].state)
        self.assertTrue(placeholder_future.done())
        self.assertEqual("on", self.machine.state_machines["my_state"]._show.name)
        self.assertEventCalled("step1_start")
        self.mock_event("step1_start")
        self.assertEventNotCalled("step1_stop")
        self.assertEventNotCalled("going_to_step2")
        self.post_event("state_machine_proceed")
        self.assertEqual("step1", self.machine.state_machines["my_state"].state)
        self.assertEventNotCalled("step1_start")
        self.assertEventNotCalled("step1_stop")
        self.assertEventNotCalled("going_to_step2")

        self.post_event("state_machine_proceed2")
        self.assertEqual("step2", self.machine.state_machines["my_state"].state)
        self.assertEqual(None, self.machine.state_machines["my_state"]._show)
        self.assertEventNotCalled("step1_start")
        self.assertEventCalled("step1_stop")
        self.assertEventCalled("going_to_step2")
        self.post_event("state_machine_proceed2")
        self.assertEqual("step2", self.machine.state_machines["my_state"].state)

        self.post_event("state_machine_proceed3")
        self.assertEqual("start", self.machine.state_machines["my_state"].state)

        self.post_event("state_machine_proceed")
        self.assertEqual("step1", self.machine.state_machines["my_state"].state)

        self.post_event("state_machine_reset")
        self.assertEqual("start", self.machine.state_machines["my_state"].state)

    def test_state_machines_in_modes(self):
        self.mock_event("non_game_mode_state_machine_done")
        self.mock_event("game_mode_state_machine_done")

        self.post_event("non_game_mode_state_machine_proceed")
        self.assertEventCalled("non_game_mode_state_machine_done")
        self.assertEventNotCalled("game_mode_state_machine_done")
        self.mock_event("non_game_mode_state_machine_done")

        self.post_event("game_mode_state_machine_proceed")
        self.assertEventNotCalled("non_game_mode_state_machine_done")
        self.assertEventNotCalled("game_mode_state_machine_done")

        self.start_game()

        self.post_event("game_mode_state_machine_proceed")
        self.assertEventNotCalled("non_game_mode_state_machine_done")
        self.assertEventCalled("game_mode_state_machine_done")

    def test_starting_state(self):
        self.assertEqual("foo", self.machine.state_machines["second_state"].state)
        self.post_event("state_machine_outoforder")
        self.assertEqual("bar", self.machine.state_machines["second_state"].state)
